import itertools
import json

import numpy as np
import pandas as pd


def load_context(path):
    return json.load(open(path))


def load_maps(path):
    def convert_inferences(infs):
        """Converts lists of antecedents to sets of antecedents."""
        return [inf | {"antecedents": set(inf["antecedents"])} for inf in infs]

    def convert_map(map_data):
        return map_data | {"inferences": convert_inferences(map_data["inferences"])}

    return [convert_map(m) for m in json.load(open(path))]


def is_na(ctx, map_data):
    return map_data["name"].split("-")[0] == ctx["action-flags"]["no-action"]


def is_a(ctx, map_data):
    return map_data["name"].split("-")[0] == ctx["action-flags"]["action"]


def is_x(ctx, map_data):
    return map_data["name"].split("-")[1] == ctx["group-flags"]["group-x"]


def is_y(ctx, map_data):
    return map_data["name"].split("-")[1] == ctx["group-flags"]["group-y"]


def is_nax(ctx, map_data):
    return is_na(ctx, map_data) and is_x(ctx, map_data)


def is_nay(ctx, map_data):
    return is_na(ctx, map_data) and is_y(ctx, map_data)


def is_ax(ctx, map_data):
    return is_a(ctx, map_data) and is_x(ctx, map_data)


def is_ay(ctx, map_data):
    return is_a(ctx, map_data) and is_y(ctx, map_data)


def get_ante(map_data):
    return [v for infe in map_data["inferences"] for v in infe["antecedents"]]


def get_cons(map_data):
    return [infe["consequent"] for infe in map_data["inferences"]]


def get_vars(map_data):
    return get_ante(map_data) + get_cons(map_data)


def get_pure_ante(map_data):
    return set(get_ante(map_data)) - set(get_cons(map_data))


def get_pure_cons(map_data):
    return set(get_cons(map_data)) - set(get_ante(map_data))


def get_all_vars(maps):
    return [v for map_data in maps for v in get_vars(map_data)]


def get_all_ante(maps):
    return [v for map_data in maps for v in get_ante(map_data)]


def get_all_cons(maps):
    return [v for map_data in maps for v in get_cons(map_data)]


def get_all_pure_ante(maps):
    return {v for map_data in maps for v in get_pure_ante(map_data)}


def get_all_pure_cons(maps):
    return {v for map_data in maps for v in get_pure_cons(map_data)}


def counterfactual(map_data, init_set, off_set):
    if not init_set:
        init_set = get_pure_ante(map_data)

    var_set = set(init_set - off_set)
    while True:
        count = len(var_set)
        for infr in map_data["inferences"]:
            if infr["antecedents"] <= var_set and infr["consequent"] not in off_set:
                var_set.add(infr["consequent"])
        if count == len(var_set):
            break
    return var_set


def counterfactual_all_maps(ctx, maps, init_set, off_set, dec_set):
    run_data = dict()
    # add counterfactual variables to output dict
    run_data.update({v: 1 for v in off_set})
    run_data["map"] = 0
    run_data["na"] = 0
    run_data["a"] = 0
    run_data["x"] = 0
    run_data["y"] = 0
    run_data["nax"] = 0
    run_data["nay"] = 0
    run_data["ax"] = 0
    run_data["ay"] = 0
    run_data["count_vars"] = len(off_set)
    for map_data in maps:
        if counterfactual(map_data, init_set, off_set) & dec_set:
            # update counters
            run_data["map"] += 1
            if is_nax(ctx, map_data):
                run_data["nax"] += 1
                run_data["na"] += 1
                run_data["x"] += 1
            if is_nay(ctx, map_data):
                run_data["nay"] += 1
                run_data["na"] += 1
                run_data["y"] += 1
            if is_ax(ctx, map_data):
                run_data["ax"] += 1
                run_data["a"] += 1
                run_data["x"] += 1
            if is_ay(ctx, map_data):
                run_data["ay"] += 1
                run_data["a"] += 1
                run_data["y"] += 1

    return run_data


def run(ctx, maps, output, mode):
    init_vars = frozenset(ctx["initial-variables"])
    env_set = frozenset(ctx["environment-variables"])
    dec_set = frozenset(ctx["decision-variables"])
    all_data = list()
    cnt = 0
    for i in range(len(env_set) + 1):
        for off_set in map(frozenset, itertools.combinations(env_set, i)):
            match mode:
                case 1:
                    init_set = init_vars | env_set
                case 2:
                    init_set = init_vars
                case _:
                    raise RuntimeError("invalid mode")
            all_data.append(
                counterfactual_all_maps(ctx, maps, init_set, off_set, dec_set)
            )
            if cnt % 1000 == 0:
                print(f"{100 * cnt / 2**len(env_set):.2f} %")
            cnt += 1
    df = pd.DataFrame(all_data)
    col_order = ["count_vars", "map", "na", "a", "x", "y", "nax", "nay", "ax", "ay"]
    col_order += sorted(env_set)
    df = df[col_order]
    df = df.fillna(0)
    df = df.astype(np.int64)
    df = df.sort_values(by=list(df.columns))
    df.to_csv(output, index=False)


run(load_context("context1.json"), load_maps("data1.json"), "output1.csv", mode=1)
run(load_context("context2.json"), load_maps("data2.json"), "output2.csv", mode=2)
